/*
Copyright (c) 2012, Felix J. Ogris
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met: 

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer. 
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/epoll.h>
#include <time.h>
#include <signal.h>
#include "fproxy.h"

/* TODO support kqueue for FreeBSD */

#define LOGFILE			"/var/log/fproxy.log"
#define ERRFILE			"/var/log/fproxy.err"
#define PIDFILE			"/var/run/fproxy.pid"
#define LISTEN_BACKLOG		1024 /* listen()                           */
#define EPOLL_HINT		1024 /* max # of events from epoll_wait()  */
#define CONN_POOL_SIZE		4096 /* max # of unused conn structs       */
#define EPOLL_TIMEOUT		1000 /* milliseconds                       */
#define PAGE_SIZE		4096 /* chunk size for allocating Strings  */
#define LOOPED_SYSCALL(A)	do { errno = 0; A; } while (errno == EINTR)
#define NULLSTRING		(struct String) { .s = NULL, .l = 0,._l = 0 }
#define NOMEM()			do { \
  write(STDERR_FILENO, __FUNCTION__, strlen(__FUNCTION__)); \
  write(STDERR_FILENO, ": no memory!\n", strlen(": no memory!\n")); \
  exit(111); \
} while(0)


enum conn_type {
  CLUSTER,  /* socket w/ listen() + bind()    */
  INCOMING, /* socket returned by accept()    */
  OUTGOING, /* socket w/ successful connect() */
  BACKEND,  /* backend machine definition     */
  DELETED,  /* marked for removal             */
  PENDING,  /* socket w/ connect() pending    */
};

/* our own string type */
struct String {
  char *s;
  uint64_t l, _l;
};

/* a connection */
struct conn {
  struct conn *next, *parent, *sibling;
  enum conn_type ct;
  int af, sock;
  socklen_t sa_len;
  char *addr, *port;
  uint64_t bytes, conns;
  time_t active_time, start_time, error_time;
  struct String buffer;
  union {
    struct sockaddr_in sai4;
    struct sockaddr_in6 sai6;
    struct sockaddr_un sun;
  } sa;
};

/* forward define */
static struct conn* get_backend_rr();

/* global variables */
static int glbl_reopen_logfiles = 0, glbl_exit_nicely = 0;
static int glbl_free_conns = 0;
static struct conn *glbl_free_list = NULL;
static unsigned int glbl_grace = 60, glbl_synproof = 0;
static struct conn* (*glbl_get_backend_func)() = get_backend_rr;
static struct conn *glbl_outgoings = NULL, *glbl_backends = NULL;

static int ultoa (const uint64_t l, char *buf) {
  static uint64_t save;
  static int len;
  static char *p;

  save = l;
  len = 0;
  p = buf;
  do {
   ++len;
   ++p;
   save /= 10;
  } while (save);
  save = l;
  do {
    --p;
    *p = '0' + save % 10;
    save /= 10;
  } while (save);
  return len;
}

/* String functions */
static struct String *str_realloc (struct String *s, uint64_t l) {
  l = ((l + PAGE_SIZE - 1) / PAGE_SIZE) * PAGE_SIZE;
  if (l == s->_l) return s;
  s->s = realloc(s->s, l);
  if (!s->s) NOMEM();
  s->_l = l;
  return s;
}

static struct String *str_cat_n (struct String *s, const char *c,
                                 const uint64_t l) {
  s = str_realloc(s, s->l + l + 1);
  memcpy(s->s + s->l, c, l);
  s->l += l;
  *(s->s + s->l) = '\0';
  return s;
}

static struct String *str_cat (struct String *s, const char *c) {
  return str_cat_n(s, c, strlen(c));
}

static struct String *str_cat_ui (struct String *s, uint64_t ui) {
  s = str_realloc(s, s->l + 21);
  s->l += ultoa(ui, s->s + s->l);
  *(s->s + s->l) = '\0';
  return s;
}

static void dt_helper2 (int now, char *buf) {
  if (now < 10) *buf++ = '0';
  ultoa(now, buf);
}

static void dt_helper (int now, int *old, char *buf) {
  if (now == *old) return;
  *old = now;
  dt_helper2(now, buf);
}

static struct String *str_cat_datetime (struct String *s) {
  static struct tm *nowtm, oldtm = {
    .tm_mday = 1,
    .tm_mon = 0,
    .tm_year = 70,
    .tm_hour = 0,
    .tm_min = 0,
    .tm_sec = 0,
    .tm_gmtoff = 0,
  };
  static long gmtoff;
  static time_t now, old = 0;
  static char tmp[] = "[01/Jan/1970:00:00:00 +0000] ";
  static const char *mon[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                               "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

  now = time(NULL);

  if (now != old) {
    nowtm = localtime(&now);
    dt_helper(nowtm->tm_mday % 100, &oldtm.tm_mday, tmp + 1);
    if (nowtm->tm_mon != oldtm.tm_mon) {
      oldtm.tm_mon = nowtm->tm_mon;
      memcpy(tmp + 4, mon[nowtm->tm_mon % 12], 3);
    }
    dt_helper((nowtm->tm_year + 1900) % 10000, &oldtm.tm_year, tmp + 8);
    dt_helper(nowtm->tm_hour % 100, &oldtm.tm_hour, tmp + 13);
    dt_helper(nowtm->tm_min % 100, &oldtm.tm_min, tmp + 16);
    dt_helper(nowtm->tm_sec % 100, &oldtm.tm_sec, tmp + 19);
    if (nowtm->tm_gmtoff != oldtm.tm_gmtoff) {
      gmtoff = nowtm->tm_gmtoff;
      oldtm.tm_gmtoff = nowtm->tm_gmtoff;
      if (gmtoff < 0) {
        tmp[22] = '-';
        gmtoff *= -1;
      } else {
        tmp[22] = '+';
      }
      gmtoff /= 60;
      dt_helper2(gmtoff % 60, tmp + 25);
      gmtoff /= 60;
      dt_helper2(gmtoff % 100, tmp + 23);
    }
  }

  return str_cat_n(s, tmp, 29);
}

/* non-fatal warnings */
#ifdef DEBUG
#  define warn(s1, s2) _warn(__FUNCTION__, __LINE__, s1, s2)
#else
#  define warn(s1, s2) _warn(s1, s2)
#endif

static void _warn (
#ifdef DEBUG
  const char *function, int line,
#endif
  char *s1, char *s2)
{
  static struct String msg = NULLSTRING;

  msg.l = 0;
  str_cat_datetime(&msg);
#ifdef DEBUG
  str_cat(&msg, function);
  str_cat_n(&msg, ":", 1);
  str_cat_ui(&msg, line);
  str_cat_n(&msg, " ", 1);
#endif
  str_cat(&msg, s1);
  if (s2) {
    str_cat_n(&msg, ": ", 2);
    str_cat(&msg, s2);
  }
  str_cat_n(&msg, "\n", 1);
  write(STDERR_FILENO, msg.s, msg.l);
}

/* fatal warnings */
#define die(s1, s2) do { \
  warn(s1, s2); \
  exit(1); \
} while (0)

static struct String *str_inet_ntop (struct String *s, int af,
                                     void *sockaddr) {
  s = str_realloc(s, s->l + 512);
  switch (af) {
    case AF_INET:
      inet_ntop(af, (void*) &((struct sockaddr_in*) sockaddr)->sin_addr,
                s->s + s->l, 512);
      break;
    case AF_INET6:
      inet_ntop(af, (void*) &((struct sockaddr_in6*) sockaddr)->sin6_addr,
                s->s + s->l, 512);
      break;
    case AF_UNIX:
      str_cat(s, ((struct sockaddr_un*) sockaddr)->sun_path);
      return s;
    default:
      die("unknown protocol", NULL);
  }
  s->l += strlen(s->s + s->l);
  return s;
}

/* connection pool */
static struct conn *new_conn (struct conn **list, int dontwarn) {
  static struct conn *newc;

  if (glbl_free_conns == 0) {
    if (!dontwarn) warn("extending connection pool", NULL);
    for (; glbl_free_conns < CONN_POOL_SIZE; ++glbl_free_conns) {
      newc = malloc(sizeof(struct conn));
      if (!newc) NOMEM();
      newc->buffer = NULLSTRING;
      newc->next = glbl_free_list;
      glbl_free_list = newc;
    }
  }
  else newc = glbl_free_list;
  glbl_free_list = newc->next;
  --glbl_free_conns;
  newc->conns = 0;
  newc->bytes = 0;
  newc->active_time = newc->start_time = time(NULL);
  newc->error_time = 0;
  newc->next = *list;
  newc->buffer.l = 0;
  *list = newc;
  return newc;
}

static void clear_list (struct conn **list) {
  static struct conn *tmpc, *prevc, *nextc;

  for (prevc = NULL, tmpc = *list; tmpc; tmpc = nextc) {
    nextc = tmpc->next;
    if (tmpc->ct == DELETED) {
      if (prevc) prevc->next = nextc;
      else *list = nextc;
      tmpc->next = glbl_free_list;
      glbl_free_list = tmpc;
      ++glbl_free_conns;
    }
    else
      prevc = tmpc;
  }
}

static void gc () {
  static struct conn *delc;

  if (glbl_free_conns >= CONN_POOL_SIZE) {
    warn("vacuuming connection pool", NULL);
    for (; glbl_free_conns > 0; --glbl_free_conns) {
      delc = glbl_free_list;
      glbl_free_list = delc->next;
      if (delc->buffer.s) free(delc->buffer.s);
      free(delc);
    }
  }
}

static void log_conn(struct conn *incom, struct conn *outgo, char *why) {
  static struct String s = NULLSTRING;
  static time_t now;

  s.l = 0;
  now = time(NULL);
  str_inet_ntop(&s, incom->af, (void*) &incom->sa);
  str_cat_n(&s, " - - ", 5);
  str_cat_datetime(&s);
  str_cat_n(&s, "\"", 1);
  if (glbl_synproof) str_cat_n(&s, "POST ", 5);
  else str_cat_n(&s, "GET ", 4);
  if (outgo) str_inet_ntop(&s, outgo->parent->af, (void*) &outgo->parent->sa);
  else str_cat_n(&s, "-", 1);
  str_cat_n(&s, "\" ", 2);
  str_cat(&s, why);
  str_cat_n(&s, " ", 1);
  str_cat_ui(&s, incom->bytes);
  str_cat_n(&s, " ", 1);
  str_cat_ui(&s, now - incom->start_time);
  if (outgo) {
    str_cat_n(&s, " ", 1);
    str_cat_ui(&s, outgo->bytes);
    str_cat_n(&s, " ", 1);
    str_cat_ui(&s, now - outgo->start_time);
    str_cat_n(&s, "\n", 1);
  }
  else str_cat_n(&s, " - -\n", 5);
  /*
   * remotehost rfc931(ident) authuser [date] "request" status bytes referer user-agent
   */
  write(STDOUT_FILENO, s.s, s.l);
}

/* close connection and unregister socket from epoll */
static void real_shutdown_conn (struct conn *delc, int epfd) {
  static int err;
  if (delc->ct != DELETED) {
    LOOPED_SYSCALL(err = epoll_ctl(epfd, EPOLL_CTL_DEL, delc->sock, NULL));
    if (err < 0) die("epoll_ctl()", strerror(errno));
    LOOPED_SYSCALL(close(delc->sock));
    delc->ct = DELETED;
  }
}

/* close bi-directional connection and make httpd compliant log entry */
static void shutdown_conn (struct conn *delc, int epfd, char *why) {
  if (delc->ct == INCOMING) log_conn(delc, delc->sibling, why);
  else log_conn(delc->sibling, delc, why);

  real_shutdown_conn(delc, epfd);

  if (delc->sibling) real_shutdown_conn(delc->sibling, epfd);
}

/* make connection to backend as given by schedule function */
static int new_backend_conn (struct conn *sibling, int epfd) {
  static struct conn *newc, *parent;
  static struct epoll_event epev;
  static int err, saved_err, flag;
  static time_t now, last_warn = 0;

  parent = glbl_get_backend_func();
  if (!parent) {
    now = time(NULL);
    if (last_warn + glbl_grace <= now) {
      warn("no more backends!", NULL);
      last_warn = now;
    }
    shutdown_conn(sibling, epfd, "503");
    return -1;
  }
  newc = new_conn(&glbl_outgoings, 0);
  newc->parent = parent;
  newc->sibling = sibling;
  sibling->sibling = newc;
  newc->ct = PENDING;
  newc->af = parent->af;
  newc->sa_len = parent->sa_len;
  LOOPED_SYSCALL(newc->sock = socket(newc->af, SOCK_STREAM, 0));
  if (newc->sock < 0) die("socket()", strerror(errno));
  LOOPED_SYSCALL(err = fcntl(newc->sock, F_GETFL, 0));
  if (err < 0) die("fcntl()", strerror(errno));
  flag = err | O_NONBLOCK;
  LOOPED_SYSCALL(err = fcntl(newc->sock, F_SETFL, flag));
  if (err < 0) die("fcntl()", strerror(errno));
  LOOPED_SYSCALL(flag = connect(newc->sock, (struct sockaddr*) &parent->sa,
                                parent->sa_len));
  saved_err = errno;
  epev.events = EPOLLIN | EPOLLPRI | EPOLLOUT;
  epev.data.ptr = (void*) newc;
  LOOPED_SYSCALL(err = epoll_ctl(epfd, EPOLL_CTL_ADD, newc->sock, &epev));
  if (err < 0) die("epoll_ctl()", strerror(errno));
  if ((saved_err != 0) && (saved_err != EINPROGRESS)) {
    warn("connect()", strerror(saved_err));
    shutdown_conn(newc, epfd, "503");
  }
  else ++parent->conns;
  return 0;
}

/* returns NULL if connection is still in error condition */
static struct conn *valid_backend (struct conn *con)
{
  if (con->error_time + glbl_grace > time(NULL)) return NULL;

  if (con->error_time != 0) {
    warn(con->addr, "re-enabling backend address");
    con->error_time = 0;
  }

  return con;
}

/* schedulers */
static struct conn *get_backend_rr () {
  static struct conn *ret = NULL;
  struct conn *start;

  if (ret) ret = ret->next;
  if (!ret) ret = glbl_backends;
  start = ret;
  while (!valid_backend(ret)) {
    ret = ret->next;
    if (!ret) ret = glbl_backends;
    if (ret == start) return NULL;
  }

  return ret;
}

static struct conn *get_backend_minconns () {
  static struct conn *ret, *tmp;

  ret = glbl_backends;
  for (tmp = ret; tmp != NULL; tmp = tmp->next)
    if (valid_backend(tmp) && (tmp->conns < ret->conns)) ret = tmp;

  return valid_backend(ret);
}

static struct conn *get_backend_minbytes () {
  static struct conn *ret, *tmp;

  ret = glbl_backends;
  for (tmp = ret; tmp != NULL; tmp = tmp->next)
    if (valid_backend(tmp) && (tmp->bytes < ret->bytes)) ret = tmp;

  return valid_backend(ret);
}

static struct conn *get_backend_mintimes () {
  static struct conn *ret, *tmp;

  ret = glbl_backends;
  for (tmp = ret; tmp != NULL; tmp = tmp->next)
    if (valid_backend(tmp) && (tmp->active_time < ret->active_time)) ret = tmp;

  return valid_backend(ret);
}

static unsigned long long algo_weighted (struct conn *con) {
  return con->conns + con->bytes / 1024 + con->active_time;
}

static struct conn *get_backend_weighted () {
  static struct conn *ret, *tmp;
  static unsigned long long old, new;

  ret = glbl_backends;
  old = algo_weighted(ret);

  for (tmp = ret; tmp != NULL; tmp = tmp->next) {
    if (valid_backend(ret)) {
      new = algo_weighted(tmp);
      if (new < old) {
        ret = tmp;
        old = new;
      }
    }
  }

  return valid_backend(ret);
}

/* wrapper for sigaction */
static void signal_set (int sig, void (*handler)(int)) {
  static struct sigaction sigact;
  static int err;

  memset(&sigact, 0, sizeof(struct sigaction));
  sigact.sa_handler = handler;
  LOOPED_SYSCALL(err = sigaction(sig, &sigact, NULL));
  if (err < 0) die("sigaction()", strerror(errno));
}

/* signal handling */
static void sighup_handler (int i) {
  i = i;
  glbl_reopen_logfiles = 1;
}

static void sigterm_handler (int i) {
  i = i;
  glbl_exit_nicely = 1;
}

/* exchange underlying file for descriptor */
static void reopen (char *fn, int fd) {
  static int nfd, err;
  LOOPED_SYSCALL(nfd = open(fn, O_CREAT | O_APPEND | O_WRONLY,
                            S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH));
  if (nfd < 0) {
    warn(fn, strerror(errno));
    return;
  }
  LOOPED_SYSCALL(err = dup2(nfd, fd));
  if (err < 0) die("dup2()", strerror(errno)); /* maybe not visible */
  LOOPED_SYSCALL(close(nfd));
}

/* show some help */
static void usage () {
  static const char *help =
"fproxy V" VERSION " - non-forking tcp proxy using the epoll() framework\n\n"
"fproxy [-o<logfile>] [-e<errorlog>] [-p<pidfile>] [-f|-ff] [-y<# bytes>]\n"
"       [-t<timeout>] [-g<grace>] [-s0] [-s1] [-s2] [-s3] [-s4]\n"
"       <cluster[:port]> [ ... ] -- <backend[:port]> [ ... ]\n\n"
"fproxy -h\n\n"
"  -h               this help ;-)\n"
"  -o<logfile>      use <logfile> for logging instead of /var/log/fproxy.log\n"
"  -e<errorlog>     use <errorlog> for error messages\n"
"                   instead of /var/log/fproxy.err\n"
"  -p<pidfile>      write process id into <pidfile>,\n"
"                   not into /var/run/fproxy.pid\n"
"  -f               do not fork and detach from controlling terminal;\n"
"                   do not use logfile, errorlog and pidfile\n"
"  -ff              do not fork and detach from controlling terminal;\n"
"                   do not use pidfile\n"
"  -y<# bytes>      open new connection to backend servers only if the "
"client\n"
"                   wrote <# bytes>, eg. use -y6 for preventing a SYN attack\n"
"                   against a webserver (nb. strlen(\"GET /\\n\") = 6);\n"
"                   -y0 disables this feature, which is also the default\n"
"  -t<timeout>      drop idle client connections after <timeout> seconds;\n"
"                   -t0 disables timeout handling, which is also the default\n"
"  -g<grace>        reenables a failed backend after <grace> seconds;\n"
"                   defaults to 60 seconds\n"
"  -s0              select backends in a round robin fashion\n"
"  -s1              select backend with least connections handled so far\n"
"  -s2              select backend with least bytes (in + out) processed so "
"far\n"
"  -s3              select backend with least connection time consumed so "
"far\n"
"  -s4              (default) select backend in a weighted fashion, using "
"number\n"
"                   of connections, processed bytes and connection time to\n"
"                   calculate to workload of each backend\n"
"  <cluster[:port]> path to Unix socket or IPv4 or IPv6 address(es) w/ "
"optional\n"
"                   tcp port on which this machine accepts connections;\n"
"                   port defaults to 80\n"
"  <backend[:port]> path to Unix socket or IPv4 or IPv6 address(es) w/ "
"optional\n"
"                   tcp port to which client connections be proxied;\n"
"                   port defaults to 80\n\n"
" send SIGHUP  to reopen logfile and errorlog\n"
" send SIGTERM to exit nicely (dont accept any new client connections, wait\n"
"              for active connections to be finished normally)\n"
" send SIGKILL to exit immediatelly\n\n";
  write(STDOUT_FILENO, help, strlen(help));
}

/* main routine */
int main (int argc, char **argv) {
  enum conn_type ct;
  struct conn *newc, *tmpc,
    *clusters = NULL, *incomings = NULL;
  int err, epfd, epevs, reuseaddr = 1;
  unsigned int dontfork = 0, timeout = 0;
  ssize_t len;
  struct epoll_event epev[EPOLL_HINT];
  time_t now;
  char *logfile = LOGFILE, *errfile = ERRFILE, *pidfile = PIDFILE,
    buffer[PAGE_SIZE];
  sigset_t sigset;

  epfd = epoll_create(EPOLL_HINT);
  if (epfd < 0) die("epoll_create()", strerror(errno));

  /* evaluate arguments */
  ct = BACKEND;
  for (--argc; argc > 0; --argc) {
    if (!strcmp(argv[argc], "--")) {
      ct = CLUSTER;
      continue;
    }
    if (!strncmp(argv[argc], "-o", 2)) {
      logfile = argv[argc] + 2;
      continue;
    }
    if (!strncmp(argv[argc], "-e", 2)) {
      errfile = argv[argc] + 2;
      continue;
    }
    if (!strncmp(argv[argc], "-p", 2)) {
      pidfile = argv[argc] + 2;
      continue;
    }
    if (!strcmp(argv[argc], "-f")) {
      if (dontfork) die("-f and -ff are mutually exclusive", NULL);
      dontfork = 1;
      continue;
    }
    if (!strcmp(argv[argc], "-ff")) {
      if (dontfork) die("-f and -ff are mutually exclusive", NULL);
      dontfork = 2;
      continue;
    }
    if (!strncmp(argv[argc], "-y", 2)) {
      glbl_synproof = atoi(argv[argc] + 2);
      ultoa(glbl_synproof, buffer);
      warn("synproof", buffer);
      continue;
    }
    if (!strncmp(argv[argc], "-t", 2)) {
      timeout = atoi(argv[argc] + 2);
      ultoa(timeout, buffer);
      warn("timeout", buffer);
      continue;
    }
    if (!strncmp(argv[argc], "-g", 2)) {
      glbl_grace = atoi(argv[argc] + 2);
      ultoa(glbl_grace, buffer);
      warn("grace", buffer);
      continue;
    }
    if (!strcmp(argv[argc], "-h")) {
      usage();
      return 0;
    }
    if (!strcmp(argv[argc], "-s0")) {
      glbl_get_backend_func = get_backend_rr;
      continue;
    }
    if (!strcmp(argv[argc], "-s1")) {
      glbl_get_backend_func = get_backend_minconns;
      continue;
    }
    if (!strcmp(argv[argc], "-s2")) {
      glbl_get_backend_func = get_backend_minbytes;
      continue;
    }
    if (!strcmp(argv[argc], "-s3")) {
      glbl_get_backend_func = get_backend_mintimes;
      continue;
    }
    if (!strcmp(argv[argc], "-s4")) {
      glbl_get_backend_func = get_backend_weighted;
      continue;
    }

    newc = new_conn((ct == BACKEND)?(&glbl_backends):(&clusters), 1);
    newc->ct = ct;
    newc->addr = argv[argc];
    newc->port = strrchr(newc->addr, ':');
    if (newc->port) *newc->port++ = '\0';
    else newc->port = "80";

    if (argv[argc][0] == '/') {
      newc->af = AF_UNIX;
      newc->sa_len = sizeof(struct sockaddr_un);
      memcpy(&newc->sa.sun.sun_path, argv[argc], strlen(argv[argc]) + 1);
      newc->sa.sun.sun_family = AF_UNIX;
    }
    else if (inet_pton(AF_INET, argv[argc], (void*) &newc->sa.sai4.sin_addr)) {
      newc->af = AF_INET;
      newc->sa_len = sizeof(struct sockaddr_in);
      newc->sa.sai4.sin_port = htons(atoi(newc->port));
      newc->sa.sai4.sin_family = AF_INET;
    }
    else if (inet_pton(AF_INET6, argv[argc],
                       (void*) &newc->sa.sai6.sin6_addr)) {
      newc->af = AF_INET6;
      newc->sa_len = sizeof(struct sockaddr_in6);
      newc->sa.sai6.sin6_port = htons(atoi(newc->port));
      newc->sa.sai6.sin6_family = AF_INET6;
    }
    else die("invalid address", argv[argc]);

    if (ct != CLUSTER) continue;

    for (tmpc = glbl_backends; tmpc; tmpc = tmpc->next)
      if (!memcmp((void*) &tmpc->sa, (void*) &newc->sa, newc->sa_len) &&
          (tmpc->sa_len == newc->sa_len))
        die("backend <-> cluster loop detected", argv[argc]);

    newc->sock = socket(newc->af, SOCK_STREAM, 0);
    if (newc->sock < 0) die("socket()", strerror(errno));
    if (setsockopt(newc->sock, SOL_SOCKET, SO_REUSEADDR, &reuseaddr,
                   sizeof(reuseaddr)) != 0)
      die("setsockopt()", strerror(errno));
    if (bind(newc->sock, (struct sockaddr*) &newc->sa, newc->sa_len) < 0)
      die("bind()", strerror(errno));
    if (listen(newc->sock, LISTEN_BACKLOG) < 0)
      die("listen()", strerror(errno));
    epev[0].events = EPOLLIN | EPOLLPRI;
    epev[0].data.ptr = (void*) newc;
    if (epoll_ctl(epfd, EPOLL_CTL_ADD, newc->sock, &epev[0]) < 0)
      die("epoll_ctl()", strerror(errno));
  }

  if (clusters == NULL) die("no clusters!", NULL);
  if (glbl_backends == NULL) die("no backends!", NULL);

  if (sigfillset(&sigset) != 0) die("sigfillset()", strerror(errno));
  if (sigdelset(&sigset, SIGHUP) != 0)
    die("sigdelset(SIGHUP)", strerror(errno));
  if (sigdelset(&sigset, SIGTERM) != 0)
    die("sigdelset(SIGTERM)", strerror(errno));
  if (sigprocmask(SIG_SETMASK, &sigset, NULL) != 0)
    die("sigprocmask()", strerror(errno));
  
  signal_set(SIGHUP, sighup_handler);
  signal_set(SIGTERM, sigterm_handler);

  if (dontfork != 1) {
    /* -ff given, or neither -f nor -ff given */
    reopen(logfile, 1);
    reopen(errfile, 2);
  }

  if (!dontfork) {
    /* daemonize */
    pid_t pid;
    pid = fork();
    if (pid < 0) die("fork()", strerror(errno));
    if (pid != 0) return 0;
    umask(0022);
    if (chdir("/") < 0) die("chdir()", strerror(errno));
    close(0);
    if (setsid() < 0) die("setsid()", strerror(errno));
    LOOPED_SYSCALL(err = open(pidfile, O_CREAT | O_TRUNC | O_WRONLY,
                              S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH));
    if (err < 0) die(pidfile, strerror(errno));
    epevs = ultoa(getpid(), buffer);
    len = write(err, buffer, epevs);
    if (len < 0) die("write()", strerror(errno));
    if (len != epevs) die(pidfile, "cannot write all data!?");
    close(err);
  }

  /* enter loop */
  warn("fproxy", "running");
  while ((glbl_exit_nicely != 2) || incomings) {
    epevs = epoll_wait(epfd, epev, EPOLL_HINT, EPOLL_TIMEOUT);
    if ((epevs < 0) && (errno != EINTR)) die("epoll_wait()", strerror(errno));
    if (epevs > EPOLL_HINT)
      die("epoll_wait()", "returned more than EPOLL_HINT events!?!");
    now = time(NULL);
    for (--epevs; epevs >= 0; --epevs) {
      tmpc = (struct conn*) epev[epevs].data.ptr;
      switch (tmpc->ct) {
        case CLUSTER:
          newc = new_conn(&incomings, 0);
          newc->parent = tmpc;
          newc->sibling = NULL;
          newc->ct = INCOMING;
          newc->af = tmpc->af;
          newc->sa_len = tmpc->sa_len;
          LOOPED_SYSCALL(newc->sock = accept(tmpc->sock,
                                             (struct sockaddr*) &newc->sa,
                                             &newc->sa_len));
          if (newc->sock < 0) {
            warn("accept()", strerror(errno));
            warn(tmpc->addr, "shutting down cluster address due to accept() "
                             "error");
            real_shutdown_conn(tmpc, epfd);
            newc->ct = DELETED;
            break;
          }
          ++tmpc->conns;
          epev[epevs].events = EPOLLIN | EPOLLPRI;
          epev[epevs].data.ptr = (void*) newc;
          LOOPED_SYSCALL(err = epoll_ctl(epfd, EPOLL_CTL_ADD, newc->sock,
                                         &epev[epevs]));
          if (err < 0) die("epoll_ctl()", strerror(errno));
          if (!glbl_synproof)
            new_backend_conn(newc, epfd);
          break;
        case INCOMING:
        case OUTGOING:
          tmpc->parent->active_time += now - tmpc->active_time;
          tmpc->active_time = now;
          if ((epev[epevs].events & EPOLLHUP) ||
              (epev[epevs].events & EPOLLERR)) {
            shutdown_conn(tmpc, epfd, "503");
            break;
          }
          if ((epev[epevs].events & EPOLLIN) ||
              (epev[epevs].events & EPOLLPRI)) {
            str_realloc(&tmpc->buffer, tmpc->buffer.l + PAGE_SIZE);
            LOOPED_SYSCALL(len = recv(tmpc->sock,
                                      tmpc->buffer.s + tmpc->buffer.l,
                                      PAGE_SIZE, MSG_DONTWAIT));
            if (len < 0) {
              warn("recv()", strerror(errno));
              shutdown_conn(tmpc, epfd, "500");
              break;
            }
            if (len == 0) {
              shutdown_conn(tmpc, epfd, "200");
              break;
            }
            tmpc->buffer.l += len;
            tmpc->bytes += len;
            tmpc->parent->bytes += len;
            if (!tmpc->sibling) {
              if (tmpc->buffer.l < glbl_synproof) break;
              if (new_backend_conn(tmpc, epfd) < 0)
                break;
            }
            if (tmpc->sibling->ct == PENDING) break;
            LOOPED_SYSCALL(len = send(tmpc->sibling->sock, tmpc->buffer.s,
                                      tmpc->buffer.l,
                                      MSG_DONTWAIT | MSG_NOSIGNAL));
            if (len < 0) {
              if (errno == EAGAIN) break;
              if (errno == EWOULDBLOCK) break;
              warn("send()", strerror(errno));
              shutdown_conn(tmpc->sibling, epfd, "500");
              break;
            }
            tmpc->buffer.l -= len;
            tmpc->sibling->active_time = now;
            memmove(tmpc->buffer.s, tmpc->buffer.s + len, tmpc->buffer.l);
          }
          break;
        case PENDING:
          if ((epev[epevs].events & EPOLLHUP) ||
              (epev[epevs].events & EPOLLERR)) {
            warn(tmpc->parent->addr, "temporarily disabling backend address "
                                     "due to connect() error");
            tmpc->parent->error_time = now;
            real_shutdown_conn(tmpc, epfd);
            new_backend_conn(tmpc->sibling, epfd);
            break;
          }
          tmpc->ct = OUTGOING;
          epev[epevs].events = EPOLLIN | EPOLLPRI;
          LOOPED_SYSCALL(err = epoll_ctl(epfd, EPOLL_CTL_MOD, tmpc->sock,
                                         &epev[epevs]));
          if (err < 0) die("epoll_ctl()", strerror(errno));
          LOOPED_SYSCALL(len = send(tmpc->sock, tmpc->sibling->buffer.s,
                                    tmpc->sibling->buffer.l,
                                    MSG_DONTWAIT | MSG_NOSIGNAL));
          if (len < 0) {
            if (errno == EAGAIN) break;
            if (errno == EWOULDBLOCK) break;
            warn("send()", strerror(errno));
            shutdown_conn(tmpc, epfd, "500");
            break;
          }
          tmpc->sibling->buffer.l -= len;
          memmove(tmpc->sibling->buffer.s, tmpc->sibling->buffer.s + len,
                  tmpc->sibling->buffer.l);
          break;
        case DELETED:
#ifdef DEBUG
          ultoa(epev[epevs].events, buffer);
          warn("got event on DELETED connection", buffer);
#endif
          break;
        case BACKEND:
          ultoa(epev[epevs].events, buffer);
          warn("got event on BACKEND definition", buffer);
          break;
      } /* switch */
    } /* for */

    /* remove timed out connections */
    if (timeout != 0) for (tmpc = incomings; tmpc; tmpc = tmpc->next)
      if (now - tmpc->active_time > timeout)
        shutdown_conn(tmpc, epfd, "502");

    /* sighup? */
    if (glbl_reopen_logfiles == 1) {
      if (dontfork == 1)
        warn("SIGHUP received", "cannot reopen logfiles in non-server mode");
      else {
        reopen(logfile, 1);
        reopen(errfile, 2);
      }
      glbl_reopen_logfiles = 0;
    }

    /* sigkill? */
    if (glbl_exit_nicely == 1) {
      if (!dontfork) unlink(pidfile);
      warn("SIGKILL received", "waiting for connections to finish");
      for (; clusters; clusters = clusters->next) {
        if (clusters->ct == DELETED) continue;
        LOOPED_SYSCALL(err = epoll_ctl(epfd, EPOLL_CTL_DEL,
                                       clusters->sock, NULL));
        if (err < 0) die("epoll_ctl()", strerror(errno));
        LOOPED_SYSCALL(close(clusters->sock));
        if (*clusters->addr == '/') LOOPED_SYSCALL(unlink(clusters->addr));
      }
      glbl_exit_nicely = 2;
    }

    /* cleanup */
    clear_list(&incomings);
    clear_list(&glbl_outgoings);
    gc();
  } /* while */

  warn("no incomings any more", "going down");
  return 0;
}
